import sys
from collections import deque
from os.path import basename

from box import Box
from tqdm import tqdm

import kgd.constants as cnst

from kgd.cli import parse_args
from kgd.common import ParseFilesManager
from kgd.constants import SERVER_IS_DOWN, PROLOGUE, STATUS_EXPLANATION
from kgd.api import KgdTaxPaymentParser
from kgd.utils import is_server_up


def main():
    args = parse_args()
    p = Box(vars(args))

    if not is_server_up(KgdTaxPaymentParser.host):
        print(SERVER_IS_DOWN)
        sys.exit(-1)

    fm = ParseFilesManager(p.fpath)
    pr = KgdTaxPaymentParser(p.token, p.timeout)

    print(PROLOGUE)
    bins = deque(fm.load_ids())

    # print_status_map()
    print(STATUS_EXPLANATION)
    with tqdm(total=fm.all_count) as pbar:
        pbar.update(fm.prs_count)

        while bins:
            incr = 1
            r = False
            if pr.failed:
                _bin = pr.failed.popleft()
                # inform tqdm about reprocessing
                incr = 0
                r = True
            else:
                _bin = bins.popleft()

            # refresh tqdm
            fname = basename(fm.curr_file)
            status = pr.status(pr.stat, _bin, fname, r)
            pbar.set_description(status)
            try:
                r = pr.process_id(_bin, p.date_range, fm.curr_file, fm.parsed_file)
                if r < 0:
                    print(SERVER_IS_DOWN)
                    exit()

            except KeyboardInterrupt:
                sys.exit(1)

            except Exception as e:
                print(e)
                sys.exit(-1)

            pbar.update(incr)


if __name__ == "__main__":
    main()




SERVER_IS_DOWN = 'Address is unreachable. Exiting...'


PROLOGUE = """
There could be parsed BINs.
It may take some time. Building list ..."""

STATUS_EXPLANATION = """
> rqe - count of BINs we got KGD API error with. Those BINs are not supposed to be reprocessed.
> rse - count of occurrences when we got bad response(KGD requests count limitation, 
html(from proxy e.g. squid) or some other trash not xml formatted). Those BINs are supposed to be reprocessed.
> se - count of occurrences when we failed(connection, network, 500, etc ).
Those BINs are supposed to be reprocessed. 
> s - count of BINs we successfully processed
> R - indicates reprocessing
>> All these indicators are actual for only this launch! 
"""


import os
from dataclasses import dataclass

from kgd.utils import load_lines, get_base_fpath, append_file


@dataclass
class FileInfo:
    fpath: str
    size: int
    num: int


def check_id(_id):
    if len(_id) != 12:
        return False
    else:
        return all([c.isdigit() for c in _id])


class ParseFilesManager:

    """ Helper for handling all
        input and output files manipulations
    """
    def __init__(self, ids_fpath, limit_fsize=100, ext='csv'):

        # TODO limit_fsize must be in MB
        self._output_fpaths = []
        self._ids_fpath = ids_fpath
        self._limit_fsize = limit_fsize
        self._ext = ext
        self._all_cnt = 0
        self._prs_cnt = 0

        # file dirname and file name without extension
        self._fpath_base = get_base_fpath(ids_fpath)

        # file path for logging parsed ids
        self._parsed_fpath = f'{self._fpath_base}.prs'

        self._prepare()

    @property
    def all_count(self):
        return self._all_cnt

    @property
    def prs_count(self):
        return self._prs_cnt

    def _prepare(self):

        # first out file
        output_fpath = f'{self._fpath_base}_out.{self._ext}'
        num = 2

        # gather all existed out files
        while os.path.exists(output_fpath):
            f_info = FileInfo(fpath=output_fpath,
                              size=os.path.getsize(output_fpath),
                              num=num - 1)
            self._output_fpaths.append(f_info)
            output_fpath = f'{self._fpath_base}_out_{num}.{self._ext}'
            num += 1

        # if we haven't created any out file
        # initialize first out file
        if not self._output_fpaths:
            open(output_fpath, 'a').close()
            f_info = FileInfo(fpath=output_fpath,
                              size=os.path.getsize(output_fpath),
                              num=num - 1)
            self._output_fpaths.append(f_info)

    def _curr(self):

        # current out file will always
        # be last item of _output_fpaths
        return self._output_fpaths[-1]

    @property
    def curr_file(self):
        return self._curr().fpath

    @property
    def parsed_file(self):
        return self._parsed_fpath

    @property
    def src_ids(self):
        return load_lines(self._ids_fpath)

    @property
    def prs_ids(self):
        return load_lines(self.parsed_file)

    def update(self, size):
        # TODO this method is never used. use it
        curr = self._curr()

        # check if size of current out file
        # more than given limitation
        if curr.size >= self._limit_fsize:
            # increase number suffix for filename
            num = curr.num+1

            # build filename
            _new_fpath = f'{self._fpath_base}_out_{num}.{self._ext}'

            # create empty out file
            open(_new_fpath, 'a').close()

            # initialization info for new out file
            f_info = FileInfo(fpath=_new_fpath, size=size, num=num)
            self._output_fpaths.append(f_info)
        else:

            # just increase current file size
            curr.size += size

    def load_ids(self):
        _ids = []

        # if there are parsed IDs
        if os.path.exists(self._parsed_fpath):
            _ids = self.prs_ids
            self._prs_cnt = len(_ids)

        # check each ID, it could be trash
        ids = [_id for _id in self.src_ids if check_id(_id)]
        self._all_cnt = len(ids)

        # exclude parsed
        if _ids:
            s = set(ids)
            s.difference_update(set(_ids))
            return list(s)
        else:
            return ids













import os
import attr
import urllib3
from collections import deque, Counter
from http.client import responses
from time import sleep

import requests
from requests import HTTPError, ConnectionError
from box import Box
from xmltodict import parse
from xml.parsers.expat import ExpatError

# from common import ParseFilesManager
from kgd.utils import is_server_up, append_file, read_file

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)


class KgdResponseError(Exception):
    """ Connection, network, 5XX errors, etc"""
    pass


class KgdRequestError(Exception):
    """ KGD requests count limitation,
    html or some other trash not xml formatted"""
    pass


def str_corrector(value):
    return value.rstrip().replace('"', "'").replace('\n', '')


def date_corrector(value):
    # return common_corrector(value).split('+')[0]
    return value.split('+')[0]


@attr.s
class PaymentData:
    """ Wrap structure provides convenient way to handle payment data.
      Such as validating, converting etc"""
    bin = attr.ib(default='')
    taxorgcode = attr.ib(converter=str_corrector, default='')
    nametaxru = attr.ib(converter=str_corrector, default='')
    nametaxkz = attr.ib(converter=str_corrector, default='')
    kbk = attr.ib(converter=str_corrector, default='')
    kbknameru = attr.ib(converter=str_corrector, default='')
    kbknamekz = attr.ib(converter=str_corrector, default='')
    paynum = attr.ib(converter=str_corrector, default='')
    paytype = attr.ib(converter=str_corrector, default='')
    entrytype = attr.ib(converter=str_corrector, default='')
    receiptdate = attr.ib(converter=date_corrector, default='')
    writeoffdate = attr.ib(converter=date_corrector, default='')
    summa = attr.ib(converter=str_corrector, default='')


def prepare(row, struct):
    """ Convert dict into tuple using
    given structure(attr class, dataclass)."""

    # cast all fields name of struct in lowercase
    _p_dict = {k.lower(): v for k, v in row.items() if k.lower()}

    # wrap in struct
    data = struct(**_p_dict)

    return attr.astuple(data)


class KgdTaxPaymentParser:
    request_template = read_file(
        os.path.join(os.path.abspath(
            os.path.join(os.path.dirname(__file__))), 'request.xml'
        )
    )
    host = 'data.egov.kz'
    url_template = "https://{}/proxy2/culs_payments?token={}"
    headers = {'user-agent': 'Apache-HttpClient/4.1.1 (java 1.5)',
               'content-type': 'text/xml'}

    def __init__(self, token, timeout):
        self.token = token
        self.timeout = timeout
        self._failed_bins = deque([])
        self.stat = Counter()
        for s in ['rqe', 'rse', 'se', 's']:
            self.stat.setdefault(s, 0)
        self.session = requests.Session()
        self.session.headers.update(self.headers)

    @property
    def failed(self):
        return self._failed_bins

    def put_fail(self, _bin):
        self._failed_bins.append(_bin)

    def load(self, _bin, date_range):
        request = self.request_template.format(_bin, *date_range)
        url = self.url_template.format(self.host, self.token)

        r = requests.post(url, request, headers=self.headers, verify=False)

        status_code = r.status_code

        if status_code != 200:
            # we are doing something wrong
            if 400 <= status_code <= 499:
                print(f'{status_code} Client error : {responses[status_code]}')
                exit()
            else:
                r.raise_for_status()

        if r.text:
            try:
                d = Box(parse(r.text)).answer

            # it could be html(from squid proxy for example)
            # or some other trash
            # not xml formatted
            except ExpatError:
                raise KgdResponseError('Not XML formatted')

        else:
            # recently KGD started to send
            # empty response to limit
            # count of requests in certain time
            # earlier they've just sending TooManyRequests exception
            raise KgdResponseError('Empty response received')

        # if we get request error
        # something wrong with our xml-request
        # all errors described in KGD API docs
        if 'err' in d:
            errcode = d.err.errorcode
            raise KgdRequestError(f'Errorcode {errcode}')

        # it might be just one payment
        payments = d.payment if isinstance(d.payment, list) else [d.payment]

        # enrich each row by bin
        for p in payments:
            p.bin = _bin

        return [prepare(p, PaymentData) for p in payments]

    def process_id(self, _bin, date_range, out_fpath, prs_fpath):

        try:
            payments = self.load(_bin, date_range)

        except KgdRequestError:
            # we are done with this bin
            append_file(prs_fpath, _bin)
            self.stat['rqe'] += 1

        except KgdResponseError:
            # just mark _bin as failed and sleep
            self._failed_bins.append(_bin)
            self.stat['rse'] += 1
            sleep(self.timeout)

        except (HTTPError, ConnectionError) as e:
            self.stat['se'] += 1
            sleep(self.timeout)
            # continue if service is available
            if is_server_up(self.host):
                self._failed_bins.append(_bin)
            else:
                return -1

        else:
            # write payments to output file
            for p in payments:
                append_file(out_fpath, ';'.join(p))
            # write bin to prs file
            append_file(prs_fpath, _bin)
            self.stat['s'] += 1

        return 1

    @staticmethod
    def status(c, _bin, fname, reprocess):
        if reprocess:
            r = 'R'
        else:
            r = ''
        curr = '{} in {} {}'.format(_bin, fname, r)
        stata = ' '.join(f'{k}:{v}' for k, v in c.items())

        return curr + ' ' + stata
